/*
 *	piece.c
 *
 *	P/ECE Linux device driver
 *	Copyright (C) 2001 Naoyuki Sawa
 *
 *	* Sun Dec 30 22:10:02 JST 2001 Naoyuki Sawa
 *	- ϡ
 */
#include <linux/kernel.h>	/* ͥAPIȤΤɬ */
#include <linux/malloc.h>	/* kmalloc/kfreeȤΤɬ */
#include <linux/module.h>	/* ⥸塼벽뤿ɬ */
#include <linux/init.h>		/* ⥸塼벽뤿ɬ */
#include <linux/usb.h>		/* USB APIȤΤɬ */

#define PIECE_VENDOR_ID		0x0e19	/* P/ECEUSBID */
#define PIECE_PRODUCT_ID	0x1000	/* (Ʊ) */
#define PIECE_MINOR_BASE	128	/* MAJOR=180(USB),MINOR=128144 */
#define PIECE_MINOR_SIZE	16	/* 饯ǥХȤʤޤ */

/* P/ECE뤿ι¤ΤǤ
 */
typedef struct _PIECE {
	struct usb_device* dev;		/* USB¤ */
	int minor;			/* ޥʡֹ(0PIECE_MINOR_SIZE-1) */
	/* bulk-in */
	int bulk_in_endpoint;		/* Х륯INɥݥȥɥ쥹 */
	int bulk_in_buffer_size;	/* Х륯INХåե */
	void* bulk_in_buffer;		/* Х륯INХåե */
	/* bulk-out */
	int bulk_out_endpoint;		/* Х륯OUTɥݥȥɥ쥹 */
	int bulk_out_buffer_size;	/* Х륯OUTХåե */
	void* bulk_out_buffer;		/* Х륯OUTХåե */
} PIECE;
PIECE* piece_table[PIECE_MINOR_SIZE];	/* ƥޥʡֹбP/ECE */

/* ⥸塼Ϣ */
int  piece_init();
void piece_exit();
/* USBϢ */
void* piece_probe(struct usb_device* dev, unsigned int ifnum, const struct usb_device_id* id);
void piece_disconnect(struct usb_device* dev, void* ptr);
void piece_free(PIECE* piece);
void piece_bulk_in_callback(struct urb* urb);
void piece_bulk_out_callback(struct urb* urb);
/* ե륷ƥϢ */
int piece_open(struct inode* inode, struct file* file);
int piece_release(struct inode* inode, struct file* file);
ssize_t piece_read(struct file* file, char* buffer, size_t count, loff_t* ppos);
ssize_t piece_write(struct file* file, const char* buffer, size_t count, loff_t* ppos);

/* ⥸塼ν/˸ƤӽФߤؿޡޤ */
module_init(piece_init);
module_exit(piece_exit);

/* ΥǥХɥ饤ФݡȤUSBID */
struct usb_device_id piece_id_table[] = {
	{ USB_DEVICE(PIECE_VENDOR_ID, PIECE_PRODUCT_ID) }, /* P/ECE */
	{ } /* ü */
};

/* ե륷ƥफƤӽФؿΥơ֥ */
struct file_operations piece_fops = {
	open:		piece_open,	/* open˸ƤФޤ */
	release:	piece_release,	/* close˸ƤФޤ */
	read:		piece_read,	/* read˸ƤФޤ */
	write:		piece_write,	/* write˸ƤФޤ */
};

/* USBɥ饤Ф°򼨤縵Υơ֥ */
struct usb_driver piece_driver = {
	name:		"piece",		/* ɥ饤̾ */
	id_table:	piece_id_table,		/* USBIDơ֥ */
	probe:		piece_probe,		/* ջ˸ƤФޤ */
	disconnect:	piece_disconnect,	/* 賰˸ƤФޤ */
	fops:		&piece_fops,		/* եؿơ֥ */
	minor:		PIECE_MINOR_BASE,	/* 16ĤΥޥʡֹΥ١ */
};

/*****************************************************************************
 *	piece_init
 *****************************************************************************/

/* ⥸塼ɤȤˡ٤ƤФޤ
 */
int
piece_init()
{
	/* USBɥ饤ФϿޤ */
	usb_register(&piece_driver);

	return 0; /*  */
}

/*****************************************************************************
 *	piece_exit
 *****************************************************************************/

/* ⥸塼򥢥ɤȤˡ٤ƤФޤ
 */
void
piece_exit()
{
	/* USBɥ饤ФϿޤ */
	usb_deregister(&piece_driver);
}

/*****************************************************************************
 *	piece_probe
 *****************************************************************************/

/* 郎³줿ȤˡUSBɥ饤ФФƸƤФޤ
 * Υɥ饤ФݡȤUSBä顢Ͽޤ
 */
void*
piece_probe(struct usb_device* dev, unsigned int ifnum, const struct usb_device_id* id)
{
	int i;
	PIECE* piece;
	struct usb_interface_descriptor* interface;
	struct usb_endpoint_descriptor* endpoint;

info(__FUNCTION__);

	piece = NULL; /* 顼ΤˡNULLȤƤ */

	/* /usr/src/linux/driver/usb/usb.cɤ¤ꡢ
	 * piece_driver.id_tableUSBIDꤷƤС
	 * ʳεˤĤpiece_probeƤФ뤳ȤϤʤϤǤ
	 * ŤΥɥ饤ФǤϡɥ饤°ơ֥USB IDꤻ
	 * ǤդΥǥХդ줿Ȥ̵probeƤӽФ
	 * ɥ饤ФݡȤ뵡狼ɤprobeȽǤƤ褦Ǥ
	 * Υɥ饤ФϿȸΤǡP/ECEʳεˤĤ
	 * piece_probeƤФ뤳ȤϤʤϤʤΤǤ
	 * ǰΤḡ뤳Ȥˤޤ
	 */
	if(dev->descriptor.idVendor != PIECE_VENDOR_ID ||
	   dev->descriptor.idProduct != PIECE_PRODUCT_ID) goto EHAND;

	/* 󥿡եǥץƤޤ */
	interface = &dev->actconfig->interface[ifnum].altsetting[0];

	/* P/ECEΰݤޤ */
	piece = (PIECE*)kmalloc(sizeof(PIECE), GFP_KERNEL);
	if(piece == NULL) {
		err(__FUNCTION__ " kmalloc(piece) failed");
		goto EHAND;
	}
	memset(piece, 0, sizeof(PIECE)); /* 顼Τˡꥢ */

	/* ǥХǥץǼޤ */
	piece->dev = dev;

	/* Х륯INɥݥȤõޤ */
	for(i = 0; i < interface->bNumEndpoints; i++) {
		endpoint = &interface->endpoint[i];
		if((endpoint->bEndpointAddress & 0x80) && /* bit7=0:OUT/1:IN */
		   (endpoint->bmAttributes & 3) == 2) {   /* 2:Х륯 */
			break;
		}
	}
	if(i == interface->bNumEndpoints) {
		err(__FUNCTION__ " bulk-in not found");
		goto EHAND;
	}
	/* ɥݥȥɥ쥹Ǽޤ */
	piece->bulk_in_endpoint = endpoint->bEndpointAddress;
	/* ڥɥƱ礭ΥХåեݤޤ */
	piece->bulk_in_buffer_size = endpoint->wMaxPacketSize;
	piece->bulk_in_buffer = kmalloc(endpoint->wMaxPacketSize, GFP_KERNEL);
	if(piece->bulk_in_buffer == NULL) {
		err(__FUNCTION__ " kmalloc(bulk_in_buffer) failed");
		goto EHAND;
	}

	/* Х륯OUTɥݥȤõޤ */
	for(i = 0; i < interface->bNumEndpoints; i++) {
		endpoint = &interface->endpoint[i];
		if(!(endpoint->bEndpointAddress & 0x80) && /* bit7=0:OUT/1:IN */
		    (endpoint->bmAttributes & 3) == 2) {   /* 2:Х륯 */
			break;
		}
	}
	if(i == interface->bNumEndpoints) {
		err(__FUNCTION__ " bulk-out not found");
		goto EHAND;
	}
	/* ɥݥȥɥ쥹Ǽޤ */
	piece->bulk_out_endpoint = endpoint->bEndpointAddress;
	/* ڥɥƱ礭ΥХåեݤޤ */
	piece->bulk_out_buffer_size = endpoint->wMaxPacketSize;
	piece->bulk_out_buffer = kmalloc(endpoint->wMaxPacketSize, GFP_KERNEL);
	if(piece->bulk_out_buffer == NULL) {
		err(__FUNCTION__ " kmalloc(bulk_out_buffer) failed");
		goto EHAND;
	}

	/* ޥʡֹζ֤˳Ǽޤ */
	for(i = 0; i < PIECE_MINOR_SIZE; i++) {
		if(piece_table[i] == NULL) break;
	}
	if(i == PIECE_MINOR_SIZE) {
		err(__FUNCTION__ " too many minors");
		goto EHAND;
	}
	piece->minor = i;
	piece_table[i] = piece;

	return piece; /* <>NULL: */

EHAND:
	piece_free(piece);
	return NULL; /* <>NULL: */
}

/*****************************************************************************
 *	piece_disconnect
 *****************************************************************************/

/* 郎곰줿Ȥˡε椷Ƥɥ饤ФФƸƤФޤ
 */
void
piece_disconnect(struct usb_device* dev, void* ptr)
{
	PIECE* piece;

info(__FUNCTION__);

	piece = (PIECE*)ptr;
	piece_free(piece);
}

/*****************************************************************************
 *	piece_free
 *****************************************************************************/

/* P/ECE¤Τޤ
 */
void
piece_free(PIECE* piece)
{
	/* P/ECE¤ΤƤƤ... */
	if(piece != NULL) {
		/* ޥʡơ֥ϿƤ顢Ͽޤ */
		if(piece_table[piece->minor] == piece) piece_table[piece->minor] = NULL;

		/* Х륯žOUTХåեƤƤ顢ޤ */
		if(piece->bulk_out_buffer != NULL) kfree(piece->bulk_out_buffer);

		/* Х륯žINХåեƤƤ顢ޤ */
		if(piece->bulk_in_buffer != NULL) kfree(piece->bulk_in_buffer);

		/* P/ECE¤Τޤ */
		kfree(piece);
	}
}

/*****************************************************************************
 *	piece_open
 *****************************************************************************/

/* ǥХե(/dev/usb/piece015)򳫤Ȥ˸ƤФޤ
 */
int
piece_open(struct inode* inode, struct file* file)
{
	int i;
	PIECE* piece;

info(__FUNCTION__);

	/* iΡɤΥޥʡֹбP/ECEޤ */
	i = MINOR(inode->i_rdev) - PIECE_MINOR_BASE;
	if(i < 0 || PIECE_MINOR_SIZE <= i) return -ENODEV; /* <0: */
	piece = piece_table[i];
	if(piece == NULL) return -ENODEV; /* <0: */

	/* θνơե¤Τ˳ǼƤޤ */
	file->private_data = piece;

	return 0; /* 0: */
}

/*****************************************************************************
 *	piece_release
 *****************************************************************************/

/* ǥХեĤȤ˸ƤФޤ */
int
piece_release(struct inode* inode, struct file* file)
{
	PIECE* piece;

info(__FUNCTION__);

	/* 줿եǤ뤳Ȥǧޤ
	 * ɬפʤȻפɡץ뤬äƤΤǤäƤߤޤ
	 */
	piece = (PIECE*)file->private_data;
	if(piece == NULL) return -ENODEV; /* <0: */

	return 0; /* 0: */
}

/*****************************************************************************
 *	piece_read
 *****************************************************************************/

/* P/ECEΥС¤(ƥȤ˻) */
typedef struct _SYSTEMINFO {
	unsigned short size;		//  0 ι¤ΤΥ
	unsigned short hard_ver;	//  2 ϡɥС
	unsigned short bios_ver;	//  4 BIOSС
	unsigned short bios_date;	//  6 BIOS YY(7):MM(4):DD(5)
	unsigned long sys_clock;	//  8 ƥࡦå(Hz)
	unsigned short vdde_voltage;	// 12 VDDE()Ű(mV)
	unsigned short resv1;		// 14 ͽ
	unsigned char *sram_top;	// 16 SRAM ϥɥ쥹
	unsigned char *sram_end;	// 20 SRAM λɥ쥹+1
	unsigned char *pffs_top;	// 24 pffs ϥɥ쥹
	unsigned char *pffs_end;	// 28 pffs λɥ쥹
} SYSTEMINFO;

int
piece_read(struct file* file, char* buffer, size_t count, loff_t* ppos)
{
	int retval;
	PIECE* piece;
	char* cmd;
	SYSTEMINFO* si;

info(__FUNCTION__);

	/* ܤreadEOF֤ޤ */
	if(*ppos != 0) return 0; /* 0:EOF */

	/* ץ˳ǼP/ECE¤ΤФޤ */
	piece = (PIECE*)file->private_data;
	if(piece == NULL) return -ENODEV; /* <0: */

	/* С׵ᥳޥɤȯԤޤ
	 *	+0,1: 0
	 *	+1,1: sizeof(SYSTEMINFO)
	 */
	cmd = piece->bulk_out_buffer;
	cmd[0] = 0;
	cmd[1] = sizeof(SYSTEMINFO);
	retval = usb_bulk_msg(piece->dev,
		usb_sndbulkpipe(piece->dev, piece->bulk_out_endpoint),
		cmd, 2, NULL, HZ * 10); /* 10åॢ */
	if(retval != 0) {
		err(__FUNCTION__ " usb_bulk_msg(out) failed");
		return retval; /* <0: */
	}

	/* С¤Τޤ */
	si = piece->bulk_in_buffer;
	retval = usb_bulk_msg(piece->dev,
		usb_rcvbulkpipe(piece->dev, piece->bulk_in_endpoint),
		si, sizeof(SYSTEMINFO), NULL, HZ * 10); /* 10åॢ */
	if(retval != 0) {
		err(__FUNCTION__ " usb_bulk_msg(in) failed");
		return retval; /* <0: */
	}

	/* ɤ߹(ȸ)Сʸޤ */
	count = snprintf(buffer, count,
		"(size)      : %d\n"
		"HW version  : %d.%d\n"
		"OS version  : %d.%d(%d/%d/%d)\n"
		"System clock: %d[Hz]\n"
		"Vdde        : %d[mV]\n"
		"(reserved)  : %d\n"
		"S-RAM       : %08x-%08x\n"
		"File system : %08x-%08x\n",
		si->size,
		si->hard_ver >> 8 & 255, si->hard_ver & 255,
		si->bios_ver >> 8 & 255, si->bios_ver & 255,
		si->bios_date >> 9 & 127, si->bios_date >> 5 & 15, si->bios_date & 31,
		si->sys_clock,
		si->vdde_voltage,
		si->resv1,
		si->sram_top, si->sram_end,
		si->pffs_top, si->pffs_end);

	/* 饯ǥХǤϥե֤ʤʤƤ⤤ߤǤ
	 * readEOF֤ˡե֤ʤƤޤ
	 */
	*ppos += count;

	return count;
}

/*****************************************************************************
 *	piece_write
 *****************************************************************************/

ssize_t
piece_write(struct file* file, const char* buffer, size_t count, loff_t* ppos)
{
	PIECE* piece;

info(__FUNCTION__);

	/* ץ˳ǼP/ECE¤ΤФޤ */
	piece = (PIECE*)file->private_data;
	if(piece == NULL) return -ENODEV;

	/* TODO: ̤ */

	return count; /* ɡ0֤ȱʱ˽ʤ顣 */
}

/*****************************************************************************
 *	piece_bulk_in_callback
 *****************************************************************************/

/* ƱХ륯INԤ褦ˤʤä餳δؿޤ
 * ޤȤäƤޤ
 */
void
piece_bulk_in_callback(struct urb* urb)
{
info(__FUNCTION__);
}

/*****************************************************************************
 *	piece_bulk_out_callback
 *****************************************************************************/

/* ƱХ륯OUTԤ褦ˤʤä餳δؿޤ
 * ޤȤäƤޤ
 */
void
piece_bulk_out_callback(struct urb* urb)
{
info(__FUNCTION__);
}

